//--------------------------------------------------------------------------
// Program to pull the information out of various types of EXIF digital 
// camera files and show it in a reasonably consistent way
//
// This module handles basic Jpeg file handling
//
// Matthias Wandel,  Dec 1999 - Dec 2002 
//--------------------------------------------------------------------------
#ifdef WIN32
#pragma comment(lib, "c:\\TopfieldDisk\\DataFiles\\SDK\\TAP_SDK_DLL")
#include "TAP_Emulations.h"
#endif

#include "tap.h"
#include "dlmalloc.h"

//#include <stdio.h>
//#include <stdlib.h>
//#include <memory.h>
//#include <stdlib.h>
//#include <string.h>
//#include <time.h>
//#include <sys/stat.h>
//#include <errno.h>
//#include <ctype.h>

/*#ifdef _WIN32
    #include <process.h>
    #include <io.h>
    #include <sys/utime.h>
#else
    #include <utime.h>
    #include <sys/types.h>
    #include <unistd.h>
    #include <errno.h>
    #include <limits.h>
#endif*/

#include "jheadinc.c"

// Storage for simplified info extracted from file.
ImageInfo_t ImageInfo;


#define MAX_SECTIONS 100
static Section_t Sections[MAX_SECTIONS];
static int SectionsRead;
static int HaveAll;

#define PSEUDO_IMAGE_MARKER 0x123; // Extra value.
//--------------------------------------------------------------------------
// Get 16 bits motorola order (always) for jpeg header stuff.
//--------------------------------------------------------------------------
static int Get16m(const void * Short)
{
    return (((uchar *)Short)[0] << 8) | ((uchar *)Short)[1];
}


//--------------------------------------------------------------------------
// Process a COM marker.
// We want to print out the marker contents as legible text;
// we must guard against random junk and varying newline representations.
//--------------------------------------------------------------------------
static void process_COM (const uchar * Data, int length)
{
    int ch;
    char Comment[MAX_COMMENT+1];
    int nch;
    int a;

    nch = 0;

    if (length > MAX_COMMENT) length = MAX_COMMENT; // Truncate if it won't fit in our structure.

    for (a=2;a<length;a++){
        ch = Data[a];

        if (ch == '\r' && Data[a+1] == '\n') continue; // Remove cr followed by lf.

        if (ch >= 32 || ch == '\n' || ch == '\t'){
            Comment[nch++] = (char)ch;
        }else{
            Comment[nch++] = '?';
        }
    }

    Comment[nch] = '\0'; // Null terminate

    if (ShowTags){
        printf("COM marker comment: %s\n",Comment);
    }

    strcpy(ImageInfo.Comments,Comment);
}

 
//--------------------------------------------------------------------------
// Process a SOFn marker.  This is useful for the image dimensions
//--------------------------------------------------------------------------
static void process_SOFn (const uchar * Data, int marker)
{
    int data_precision, num_components;

    data_precision = Data[2];
    ImageInfo.Height = Get16m(Data+3);
    ImageInfo.Width = Get16m(Data+5);
    num_components = Data[7];

    if (num_components == 3){
        ImageInfo.IsColor = 1;
    }else{
        ImageInfo.IsColor = 0;
    }

    ImageInfo.Process = marker;

    if (ShowTags){
        printf("JPEG image is %uw * %uh, %d color components, %d bits per sample\n",
                   ImageInfo.Width, ImageInfo.Height, num_components, data_precision);
    }
}


int readByte(TYPE_File *infile)
{
	int i;
	unsigned char c;

	TAP_Hdd_Fread(&c, 1, 1, infile);
	i = c;

	return i;
}

//--------------------------------------------------------------------------
// Parse the marker stream until SOS or EOI is seen;
//--------------------------------------------------------------------------
int ReadJpegSections (TYPE_File *infile, ReadMode_t ReadMode)
{
    int a;
	
    int HaveCom = FALSE;

    //a = fgetc(infile);
	a = readByte(infile);

    if (a != 0xff || readByte(infile) != M_SOI){
        return FALSE;
    }
    for(;;){
        int itemlen;
        int marker = 0;
        int ll,lh, got;
        uchar * Data;

        if (SectionsRead >= MAX_SECTIONS){
            printf("Too many sections in jpg file");
        }

        for (a=0;a<7;a++){
            marker = readByte(infile);
            if (marker != 0xff) break;

            if (a >= 6){
                printf("too many padding bytes\n");
                return FALSE;
            }
        }

        if (marker == 0xff){
            // 0xff is legal padding, but if we get that many, something's wrong.
            printf("too many padding bytes!");
        }

        Sections[SectionsRead].Type = marker;
  
        // Read the length of the section.
        lh = readByte(infile);
        ll = readByte(infile);

        itemlen = (lh << 8) | ll;

        if (itemlen < 2){
            printf("invalid marker");
        }

        Sections[SectionsRead].Size = itemlen;

        Data = (uchar *)dlmalloc(itemlen);
        if (Data == NULL){
            printf("Could not allocate memory");
        }
        Sections[SectionsRead].Data = Data;

        // Store first two pre-read bytes.
        Data[0] = (uchar)lh;
        Data[1] = (uchar)ll;

        got = TAP_Hdd_Fread(Data+2, 1, itemlen-2, infile); // Read the whole section.
        if (got != itemlen-2){
            printf("Premature end of file?");
        }
        SectionsRead += 1;

        switch(marker){

            case M_SOS:   // stop before hitting compressed data 
                return TRUE;

            case M_EOI:   // in case it's a tables-only JPEG stream
                printf("No image in jpeg!\n");
                return FALSE;

            case M_COM: // Comment section
                if (HaveCom || ((ReadMode & READ_EXIF) == 0)){
                    // Discard this section.
                    dlfree(Sections[--SectionsRead].Data);
                }else{
                    process_COM(Data, itemlen);
                    HaveCom = TRUE;
                }
                break;

            case M_JFIF:
                // Regular jpegs always have this tag, exif images have the exif
                // marker instead, althogh ACDsee will write images with both markers.
                // this program will re-create this marker on absence of exif marker.
                // hence no need to keep the copy from the file.
                dlfree(Sections[--SectionsRead].Data);
                break;

            case M_EXIF:
                // Seen files from some 'U-lead' software with Vivitar scanner
                // that uses marker 31 for non exif stuff.  Thus make sure 
                // it says 'Exif' in the section before treating it as exif.
                if ((ReadMode & READ_EXIF) && memcmp(Data+2, "Exif", 4) == 0){
                    process_EXIF(Data, itemlen);
                }else{
                    // Discard this section.
                    dlfree(Sections[--SectionsRead].Data);
                }
                break;

            case M_SOF0: 
            case M_SOF1: 
            case M_SOF2: 
            case M_SOF3: 
            case M_SOF5: 
            case M_SOF6: 
            case M_SOF7: 
            case M_SOF9: 
            case M_SOF10:
            case M_SOF11:
            case M_SOF13:
            case M_SOF14:
            case M_SOF15:
                process_SOFn(Data, marker);
                break;
            default:
                // Skip any other sections.
                if (ShowTags){
                    printf("Jpeg section marker 0x%02x size %d\n",marker, itemlen);
                }
                break;
        }
    }
    return TRUE;
}

//--------------------------------------------------------------------------
// Discard read data.
//--------------------------------------------------------------------------
void DiscardData(void)
{
    int a;
    for (a=0;a<SectionsRead;a++){
        dlfree(Sections[a].Data);
    }
    memset(&ImageInfo, 0, sizeof(ImageInfo));
    SectionsRead = 0;
    HaveAll = 0;
}

//--------------------------------------------------------------------------
// Read image data.
//--------------------------------------------------------------------------
int ReadJpegFile(char * FileName, ReadMode_t ReadMode)
{
    TYPE_File *infile;
    int ret, size;
	
    //infile = fopen(FileName, "rb"); // Unix ignores 'b', windows needs it.
	infile = TAP_Hdd_Fopen(FileName);

    if (infile == NULL) {
        //fprintf(stderr, "can't open '%s'\n", FileName);
        return FALSE;
    }

	size = TAP_Hdd_Flen(infile);
	//if ((size > 0) && (size % 512) == 0) {
	//	size = size+512;
	//}

    // Scan the JPEG headers.
    ret = ReadJpegSections(infile, ReadMode);
    if (!ret){
        printf("Not JPEG: %s\n",FileName);
    }

    TAP_Hdd_Fclose(infile);

    if (ret == FALSE){
        DiscardData();
    }
    return ret;
}

//--------------------------------------------------------------------------
// Check if image has exif header.
//--------------------------------------------------------------------------
Section_t * FindSection(int SectionType)
{
    int a;

    for (a=0;a<SectionsRead;a++){
        if (Sections[a].Type == SectionType){
            return &Sections[a];
        }
    }
    // Could not be found.
    return NULL;
}

//--------------------------------------------------------------------------
// Save exif thumbnail to named file
//--------------------------------------------------------------------------
int SaveThumbnailData(char * ThumbFileName)
{
    TYPE_File *ThumbnailFile;

    if (ImageInfo.ThumbnailOffset == 0 || ImageInfo.ThumbnailSize == 0){
        printf("Image contains no thumbnail\r\n");
		DiscardData();
        return FALSE;
    }

	if (TAP_Hdd_Exist(ThumbFileName)) TAP_Hdd_Delete(ThumbFileName);
	TAP_Hdd_Create(ThumbFileName, ATTR_NORMAL);
	ThumbnailFile = TAP_Hdd_Fopen(ThumbFileName);

    if (ThumbnailFile){
        char * ThumbnailPointer;
        Section_t * ExifSection;
        ExifSection = FindSection(M_EXIF);
        ThumbnailPointer = ExifSection->Data+ImageInfo.ThumbnailOffset+8;

        TAP_Hdd_Fwrite(ThumbnailPointer, ImageInfo.ThumbnailSize, 1, ThumbnailFile);
        TAP_Hdd_Fclose(ThumbnailFile);
		DiscardData();
        return TRUE;
    }else{
        printf("Could not write thumbnail file\r\n");
		DiscardData();
        return FALSE;
    }
}

/** get the pointer to the thumbnail data (if it exists)
    can we assume that the thumbnail is a 160 x 120 JPEG? */ 
char *getThumbnailData()
{
	char * ThumbnailPointer;
	Section_t * ExifSection;

    if (ImageInfo.ThumbnailOffset == 0 || ImageInfo.ThumbnailSize == 0){
        printf("Image contains no thumbnail\r\n");
		DiscardData();
        return NULL;
    }
        
    ExifSection = FindSection(M_EXIF);
    ThumbnailPointer = (char *) ExifSection->Data+ImageInfo.ThumbnailOffset+8;

	// return pointer to the thumbnail data
    return ThumbnailPointer;
}